@@ -5,9 +5,15 @@ The application classes in this module are installed as executables via
 setuptools entry points.
 """
 
+from __future__ import absolute_import
+from __future__ import unicode_literals
+
+# import udiskie.depend first - for side effects!
+from .depend import has_Notify, has_Gtk, _in_X
+
 import inspect
-import logging
-import warnings
+import logging.config
+import traceback
 
 from docopt import docopt, DocoptExit
 
@@ -16,8 +22,9 @@ from gi.repository import GLib
 import udiskie
 import udiskie.config
 import udiskie.mount
-from udiskie.common import extend
-from udiskie.locale import _
+from .async_ import AsyncList, Coroutine, Return, RunForever
+from .common import extend, str2unicode
+from .locale import _
 
 
 __all__ = [
@@ -27,49 +34,36 @@ __all__ = [
 ]
 
 
-warnings.filterwarnings("ignore", ".*could not open display.*", Warning)
-warnings.filterwarnings("ignore", ".*g_object_unref.*", Warning)
-
-
-def get_backend(clsname, version=None):
+@Coroutine.from_generator_function
+def get_backend(version=None):
     """
     Return UDisks service.
 
-    :param str clsname: requested service object
     :param int version: requested UDisks backend version
     :returns: UDisks service wrapper object
-    :raises dbus.DBusException: if unable to connect to UDisks dbus service.
+    :raises GLib.GError: if unable to connect to UDisks dbus service.
     :raises ValueError: if the version is invalid
 
     If ``version`` has a false truth value, try to connect to UDisks1 and
     fall back to UDisks2 if not available.
     """
     if not version:
-        from udiskie.dbus import DBusException
         try:
-            return get_backend(clsname, 2)
-        except DBusException:
+            daemon = yield get_backend(2)
+        except GLib.GError:
             log = logging.getLogger(__name__)
             log.warning(_('Failed to connect UDisks2 dbus service..\n'
                           'Falling back to UDisks1.'))
-            return get_backend(clsname, 1)
+            daemon = yield get_backend(1)
     elif version == 1:
         import udiskie.udisks1
-        return getattr(udiskie.udisks1, clsname)()
+        daemon = yield udiskie.udisks1.Daemon.create()
     elif version == 2:
         import udiskie.udisks2
-        return getattr(udiskie.udisks2, clsname)()
+        daemon = yield udiskie.udisks2.Daemon.create()
     else:
         raise ValueError(_("UDisks version not supported: {0}!", version))
-
-
-def module_available(name):
-    """Check if the module is importable."""
-    try:
-        __import__(name)
-        return True
-    except ImportError:
-        return False
+    yield Return(daemon)
 
 
 class Choice(object):
@@ -110,7 +104,7 @@ class Value(object):
 
     def __call__(self, args):
         """Get the value of the command line argument."""
-        return args[self._name]
+        return str2unicode(args[self._name])
 
 
 class OptionalValue(object):
@@ -122,7 +116,14 @@ class OptionalValue(object):
 
     def __call__(self, args):
         """Get the value of the command line argument."""
-        return self._choice(args) and args[self._name]
+        return self._choice(args) and str2unicode(args[self._name])
+
+
+class SelectLevel(logging.Filter):
+    def __init__(self, level):
+        self.level = level
+    def filter(self, record):
+        return record.levelno == self.level
 
 
 class _EntryPoint(object):
@@ -170,11 +171,36 @@ class _EntryPoint(object):
         program_opts = self.program_options(args)
         # initialize logging configuration:
         log_level = program_opts.get('log_level', default_opts['log_level'])
-        if log_level <= logging.DEBUG:
-            fmt = _('%(levelname)s [%(asctime)s] %(name)s: %(message)s')
-        else:
-            fmt = _('%(message)s')
-        logging.basicConfig(level=log_level, format=fmt)
+        debug = log_level <= logging.DEBUG
+        logging.config.dictConfig({
+            'version': 1,
+            'disable_existing_loggers': False,
+            'formatters': {
+                'plain':  {'format': _('%(message)s')},
+                'detail': {'format': _('%(levelname)s [%(asctime)s] %(name)s: %(message)s')},
+            },
+            'filters': {
+                'info': {'()': 'udiskie.cli.SelectLevel', 'level': logging.INFO},
+            },
+            'handlers': {
+                'info':  {'class': 'logging.StreamHandler',
+                          'stream': 'ext://sys.stdout',
+                          'formatter': 'plain',
+                          'filters': ['info']},
+                'error': {'class': 'logging.StreamHandler',
+                          'stream': 'ext://sys.stderr',
+                          'formatter': 'plain',
+                          'level': 'WARNING'},
+                'debug': {'class': 'logging.StreamHandler',
+                          'stream': 'ext://sys.stderr',
+                          'formatter': 'detail'},
+            },
+            # configure root logger:
+            'root': {
+                'handlers': ['info', 'debug' if debug else 'error'],
+                'level': log_level,
+            },
+        })
         # parse config options
         config_file = OptionalValue('--config')(args)
         config = udiskie.config.Config.from_file(config_file)
@@ -185,7 +211,7 @@ class _EntryPoint(object):
         # initialize instance variables
         self.config = config
         self.options = options
-        self._init(config, options)
+        self.exit_status = 0
 
     def program_options(self, args):
         """
@@ -228,24 +254,43 @@ class _EntryPoint(object):
         """Get the name of the CLI utility."""
         raise NotImplementedError()
 
-    def _init(self, config, options):
+    def _init(self):
         """
         Fully initialize Daemon object.
 
-        :param Config config: configuration object
-        :param options: program options
+        :returns: the main application task
+        :rtype: Async
         """
         raise NotImplementedError()
 
     def run(self):
         """
-        Run main program logic.
+        Run the main loop.
 
-        :param options: program options
         :returns: exit code
         :rtype: int
         """
-        raise NotImplementedError()
+        self.mainloop = GLib.MainLoop()
+        self._start_async_tasks()
+        try:
+            self.mainloop.run()
+            return self.exit_status
+        except KeyboardInterrupt:
+            return 1
+
+    @Coroutine.from_generator_function
+    def _start_async_tasks(self):
+        """Start asynchronous operations."""
+        try:
+            self.udisks = yield get_backend(self.options['udisks_version'])
+            results = yield self._init()
+            if not all(results):
+                self.exit_status = 1
+        except Exception:
+            self.exit_status = 1
+            # Print the stack trace only up to the current level:
+            traceback.print_exc()
+        self.mainloop.quit()
 
 
 class Daemon(_EntryPoint):
@@ -282,6 +327,9 @@ class Daemon(_EntryPoint):
         -s, --smart-tray                        Auto hide tray icon
         -T, --no-tray                           Disable tray icon
 
+        --password-cache MINUTES                Set password cache timeout
+        --no-password-cache                     Disable password cache
+
         -p COMMAND, --password-prompt COMMAND   Command for password retrieval
         -P, --no-password-prompt                Disable unlocking
 
@@ -298,6 +346,7 @@ class Daemon(_EntryPoint):
         'tray': False,
         'file_manager': 'xdg-open',
         'password_prompt': 'builtin:gui',
+        'password_cache': False,
     })
 
     option_rules = extend(_EntryPoint.option_rules, {
@@ -309,26 +358,36 @@ class Daemon(_EntryPoint):
             '--smart-tray': 'auto'}),
         'file_manager': OptionalValue('--file-manager'),
         'password_prompt': OptionalValue('--password-prompt'),
+        'password_cache': OptionalValue('--password-cache'),
     })
 
-    def _init(self, config, options):
+    def _init(self):
 
         """Implements _EntryPoint._init."""
 
         import udiskie.prompt
 
-        mainloop = GLib.MainLoop()
-        daemon = get_backend('Daemon', options['udisks_version'])
+        config = self.config
+        options = self.options
+
         prompt = udiskie.prompt.password(options['password_prompt'])
         browser = udiskie.prompt.browser(options['file_manager'])
+        cache = None
+
+        if options['password_cache'] is not False:
+            import udiskie.cache
+            timeout = int(options['password_cache']) * 60
+            cache = udiskie.cache.PasswordCache(timeout)
+
         mounter = udiskie.mount.Mounter(
             mount_options=config.mount_options,
             ignore_device=config.ignore_device,
             prompt=prompt,
             browser=browser,
-            udisks=daemon)
+            cache=cache,
+            udisks=self.udisks)
 
-        if options['notify'] and not module_available('gi.repository.Notify'):
+        if options['notify'] and not has_Notify():
             libnotify_not_available = _(
                 "Typelib for 'libnotify' is not available. Possible causes include:"
                 "\n\t- libnotify is not installed"
@@ -338,6 +397,25 @@ class Daemon(_EntryPoint):
             logging.getLogger(__name__).error(libnotify_not_available)
             options['notify'] = False
 
+        if options['tray'] and not has_Gtk(3) and not _in_X:
+            no_X_session = _(
+                "Not run within X session. "
+                "\nStarting udiskie without tray icon.\n")
+            logging.getLogger(__name__).error(no_X_session)
+            options['tray'] = False
+
+        if options['tray'] and not has_Gtk(3):
+            gtk3_not_available = _(
+                "Typelib for 'Gtk 3.0' is not available. Possible causes include:"
+                "\n\t- GTK3 is not installed"
+                "\n\t- the typelib is provided by a separate package"
+                "\n\t- GTK3 was built with introspection disabled"
+                "\nStarting udiskie without tray icon.\n")
+            logging.getLogger(__name__).error(gtk3_not_available)
+            options['tray'] = False
+
+        tasks = []
+
         # notifications (optional):
         if options['notify']:
             import udiskie.notify
@@ -366,30 +444,25 @@ class Daemon(_EntryPoint):
                 mounter,
                 icons,
                 actions,
-                quit_action=mainloop.quit)
+                quit_action=self.mainloop.quit)
             TrayIcon = tray_classes[options['tray']]
             statusicon = TrayIcon(menu_maker, icons)
+            tasks.append(statusicon.task)
         else:
             statusicon = None
+            tasks.append(RunForever)
 
         # automounter
         if options['automount']:
             import udiskie.automount
             udiskie.automount.AutoMounter(mounter)
+            tasks.append(mounter.add_all())
 
         # Note: mounter and statusicon are saved so these are kept alive:
-        self.mainloop = mainloop
         self.mounter = mounter
         self.statusicon = statusicon
 
-    def run(self):
-        """Implements _EntryPoint.run."""
-        if self.options['automount']:
-            self.mounter.add_all()
-        try:
-            return self.mainloop.run()
-        except KeyboardInterrupt:
-            return 0
+        return AsyncList(tasks)
 
 
 class Mount(_EntryPoint):
@@ -443,35 +516,35 @@ class Mount(_EntryPoint):
         'password_prompt': OptionalValue('--password-prompt'),
     })
 
-    def _init(self, config, options):
+    def _init(self):
+
         """Implements _EntryPoint._init."""
+
         import udiskie.prompt
+
+        config = self.config
+        options = self.options
+
         if options['options']:
             opts = [o.strip() for o in options['options'].split(',')]
             mount_options = lambda dev: opts
         else:
             mount_options = config.mount_options
+
         prompt = udiskie.prompt.password(options['password_prompt'])
-        self.mounter = udiskie.mount.Mounter(
+        mounter = udiskie.mount.Mounter(
             mount_options=mount_options,
             ignore_device=config.ignore_device,
             prompt=prompt,
-            udisks=get_backend('Sniffer', options['udisks_version']))
+            udisks=self.udisks)
 
-    def run(self):
-        """Implements _EntryPoint.run."""
-        options = self.options
-        mounter = self.mounter
         recursive = options['recursive']
-        # only mount the desired devices
         if options['<device>']:
-            success = True
-            for path in options['<device>']:
-                success = success and mounter.add(path, recursive=recursive)
-        # mount all present devices
+            tasks = [mounter.add(path, recursive=recursive)
+                     for path in options['<device>']]
         else:
-            success = mounter.add_all(recursive=recursive)
-        return 0 if success else 1
+            tasks = [mounter.add_all(recursive=recursive)]
+        return AsyncList(tasks)
 
 
 class Umount(_EntryPoint):
@@ -531,23 +604,20 @@ class Umount(_EntryPoint):
         '<device>': Value('DEVICE'),
     })
 
-    def _init(self, config, options):
+    def _init(self):
+
         """Implements _EntryPoint._init."""
-        self.mounter = udiskie.mount.Mounter(
-            udisks=get_backend('Sniffer', options['udisks_version']))
 
-    def run(self):
-        """Implements _EntryPoint.run."""
         options = self.options
-        mounter = self.mounter
+        mounter = udiskie.mount.Mounter(self.udisks)
+
         strategy = dict(detach=options['detach'],
                         eject=options['eject'],
                         lock=options['lock'])
         if options['<device>']:
             strategy['force'] = options['force']
-            success = True
-            for path in options['<device>']:
-                success = mounter.remove(path, **strategy) and success
+            tasks = [mounter.remove(path, **strategy)
+                     for path in options['<device>']]
         else:
-            success = mounter.remove_all(**strategy)
-        return 0 if success else 1
+            tasks = [mounter.remove_all(**strategy)]
+        return AsyncList(tasks)
