@@ -16,6 +16,7 @@ __all__ = [
     'check_call',
     'Emitter',
     'samefile',
+    'sameuuid',
     'setdefault',
     'extend',
     'cachedproperty',
@@ -97,6 +98,11 @@ def samefile(a, b):
         return os.path.normpath(a) == os.path.normpath(b)
 
 
+def sameuuid(a, b):
+    """Compare two UUIDs."""
+    return a and b and a.lower() == b.lower()
+
+
 def setdefault(self, other):
     """
     Merge two dictionaries like .update() but don't overwrite values.
@@ -191,6 +197,7 @@ class NullDevice(object):
     is_partition = False
     is_filesystem = False
     is_luks = False
+    is_loop = False
 
     # Drive
     is_toplevel = is_drive
@@ -232,6 +239,7 @@ class NullDevice(object):
     should_automount = False
     icon_name = ''
     symbolic_icon_name = icon_name
+    symlinks = []
 
     # Partition
     @property
@@ -252,6 +260,7 @@ class NullDevice(object):
         raise RuntimeError("Cannot call methods on invalid device!")
 
     # Encrypted
+    @property
     def luks_cleartext_holder(self):
         raise AttributeError('Invalid device has no cleartext holder.')
 
@@ -263,6 +272,9 @@ class NullDevice(object):
     def lock(self):
         raise RuntimeError("Cannot call methods on invalid device!")
 
+    # Loop
+    loop_file = ''
+
     # derived properties
     in_use = False
     parent_object_path = '/'
