@@ -10,7 +10,7 @@ from __future__ import unicode_literals
 
 import logging
 import os
-import sys
+import fnmatch
 
 from .common import exc_message
 from .compat import basestring, fix_str_conversions
@@ -29,6 +29,14 @@ def lower(s):
         return s
 
 
+def match_value(value, pattern):
+    if isinstance(value, (list, tuple)):
+        return any(match_value(v, pattern) for v in value)
+    if isinstance(value, basestring) and isinstance(pattern, basestring):
+        return fnmatch.fnmatch(value.lower(), pattern.lower())
+    return lower(value) == lower(pattern)
+
+
 def yaml_load(stream):
     """Load YAML document, but load all strings as unicode on py2."""
     import yaml
@@ -52,6 +60,7 @@ class DeviceFilter(object):
         'is_partition',
         'is_filesystem',
         'is_luks',
+        'is_loop',
         'is_toplevel',
         'is_detachable',
         'is_ejectable',
@@ -74,6 +83,8 @@ class DeviceFilter(object):
         'in_use',
         'should_automount',
         'ui_label',
+        'loop_filename',
+        'symlinks',
     ]
 
     def __init__(self, match, value):
@@ -105,7 +116,7 @@ class DeviceFilter(object):
 
         :param Device device: device to be checked
         """
-        return all(lower(getattr(device, k)) == lower(v)
+        return all(match_value(getattr(device, k), v)
                    for k, v in self._match.items())
 
     def value(self, device):
@@ -167,6 +178,8 @@ class FilterMatcher(object):
         :param Device device: device to be mounted
         :returns: value of the first matching filter
         """
+        if device is None:
+            return self._default
         matches = (f.value(device) for f in self._filters if f.match(device))
         return next(matches, self._default)
 
