@@ -23,7 +23,7 @@ import os.path
 from gi.repository import GLib
 
 from .async_ import AsyncList, Coroutine, Return
-from .common import Emitter, samefile, AttrDictView, wraps, NullDevice
+from .common import Emitter, samefile, sameuuid, AttrDictView, wraps, NullDevice
 from .compat import fix_str_conversions
 from .dbus import connect_service, MethodsProxy
 from .locale import _
@@ -73,7 +73,9 @@ class Device(object):
         """Comparison by mount and device file path."""
         return (samefile(path, self.device_file) or
                 samefile(path, self.loop_file) or
-                any(samefile(path, mp) for mp in self.mount_paths))
+                any(samefile(path, mp) for mp in self.mount_paths) or
+                sameuuid(path, self.id_uuid) or
+                sameuuid(path, self.partition_uuid))
 
     # availability of interfaces
     @property
@@ -266,6 +268,16 @@ class Device(object):
 
     symbolic_icon_name = icon_name
 
+    @property
+    def symlinks(self):
+        """Known symlinks of the block device."""
+        return list(filter(None, [
+            self._P.DeviceFile,
+            self._P.DeviceFilePresentation,
+            self._P.DeviceFileById,
+            self._P.DeviceFileByPath,
+        ]))
+
     # ----------------------------------------
     # Partition
     # ----------------------------------------
@@ -278,6 +290,11 @@ class Device(object):
             return None
         return self._daemon[self._P.PartitionSlave]
 
+    @property
+    def partition_uuid(self):
+        """Get the partition UUID."""
+        return self._P.PartitionUuid
+
     # ----------------------------------------
     # Filesystem
     # ----------------------------------------
@@ -348,7 +365,7 @@ class Device(object):
         return self._M.LuksLock('(as)', [])
 
     # ----------------------------------------
-    # loop
+    # Loop
     # ----------------------------------------
 
     @property
