@@ -18,7 +18,7 @@ import logging
 
 from gi.repository import GLib
 
-from .common import Emitter, samefile, AttrDictView, decode_ay
+from .common import Emitter, samefile, sameuuid, AttrDictView, decode_ay
 from .compat import fix_str_conversions
 from .dbus import connect_service, MethodsProxy
 from .locale import _
@@ -151,7 +151,9 @@ class Device(object):
         """Comparison by mount and device file path."""
         return (samefile(path, self.device_file) or
                 samefile(path, self.loop_file) or
-                any(samefile(path, mp) for mp in self.mount_paths))
+                any(samefile(path, mp) for mp in self.mount_paths) or
+                sameuuid(path, self.id_uuid) or
+                sameuuid(path, self.partition_uuid))
 
     # availability of interfaces
     @property
@@ -392,6 +394,13 @@ class Device(object):
         """Return the recommended device symbolic icon name."""
         return self._P.Block.HintSymbolicIconName or 'drive-removable-media'
 
+    @property
+    def symlinks(self):
+        """Known symlinks of the block device."""
+        if not self._P.Block.Symlinks:
+            return []
+        return [decode_ay(path) for path in self._P.Block.Symlinks]
+
     # ----------------------------------------
     # Partition
     # ----------------------------------------
@@ -402,6 +411,11 @@ class Device(object):
         """Get the partition slave (container)."""
         return self._daemon[self._P.Partition.Table]
 
+    @property
+    def partition_uuid(self):
+        """Get the partition UUID."""
+        return self._P.Partition.UUID
+
     # ----------------------------------------
     # Filesystem
     # ----------------------------------------
@@ -483,7 +497,7 @@ class Device(object):
         )
 
     # ----------------------------------------
-    # loop
+    # Loop
     # ----------------------------------------
 
     @property
